{-| Functions to compile agdaLight files to LaTeX documents.
-}

module Compilers.Latex where

import Control.Monad.State
import Control.Monad.Writer
import Data.Generics
import Data.List

import Syntax.Internal
import Syntax.ConcreteToInternal
import Syntax.Lexer
import Syntax.Parser
import Syntax.ParseMonad
import Syntax.Position

import Interface.Options

import Utils.Monads
import Utils.Lists
import Utils.Generics

-- Mapping the internal categories of names back to the token stream ------

data IName  = IVar Range Name | IDef Range Name
	    | ISort Range Name
    deriving Show

getName :: IName -> Name
getName (IVar _ x) = x
getName (IDef _ x) = x
getName (ISort _ x) = x

setRange r (IVar _ x)	= IVar r x
setRange r (IDef _ x)	= IDef r x
setRange r (ISort _ x)	= ISort r x

instance HasRange IName where
    getRange (IVar r _)	    = r
    getRange (IDef r _)	    = r
    getRange (ISort r _)    = r

value x = case p of
	    NoPos   -> error $ show x
	    _	    -> (posLine p, posCol p)
    where
	p = rStart (getRange x)

instance Eq IName where
    x == y = (getName x, value x) == (getName y, value y)

instance Ord IName where
    compare x y = compare (value x) (value y)

internalNames :: Data a => a -> [IName]
internalNames =
    everythingBut   (++)
		    isString
		    ([] `mkQ` term `extQ` local
			`extQ` top
		    )
    where
	term :: Term -> [IName]
	term t =
	    case t of
		Var r x _   -> [IVar r x]
		Def r x _   -> [IDef r x]
		Star r 0    -> [ISort r "Set"]
		Star r 1    -> [ISort r "Type"]
		Prop r	    -> [ISort r "Prop"]
		_	    -> []

	local :: LocalDefinition -> [IName]
	local (LocalDef r x (Just _) r' e _)	= [IVar r x, IVar r' x]
	local (LocalDef r x _ _ _ _)		= [IVar r x]

	top :: Definition -> [IName]
	top (NormalDef r x _ e) = [IDef r x, IDef (getRange e) x]
	top (NewAxiom r x _)	= [IDef r x]

getInternalNames :: Program -> [IName]
getInternalNames p = 
    map head
	$ group
	$ sort
	$ filter (\x -> getName x /= "_")
	$ internalNames p

-- Fancy tokens -----------------------------------------------------------

data Space  = Box
	    | Fixed Int
    deriving Show

data FancyToken	= SimpleTok Token
		| IdTok IName
		| OpTok IName
		| SpaceTok Range Space
    deriving Show

instance HasRange FancyToken where
    getRange (SimpleTok t)  = getRange t
    getRange (IdTok x)	    = getRange x
    getRange (OpTok x)	    = getRange x
    getRange (SpaceTok r _) = r

formatTokens :: [IName] -> [Token] -> [FancyToken]
formatTokens _ [] = []
formatTokens [] ts = map SimpleTok ts
formatTokens (iname:ns) (t@(TkSet r) : ts)
    | getName iname == "Set"	= IdTok (setRange r iname) : formatTokens ns ts
    | otherwise			= formatTokens ns (t : ts)
{-    | otherwise			= error $ "ns = " ++ show (iname:ns)
				    ++ "\n" ++ replicate 50 '-'
				    ++ "\nts = " ++ show (TkSet r : ts)
-}
formatTokens (iname:ns) (t@(TkProp r) : ts)
    | getName iname == "Prop"	= IdTok (setRange r iname) : formatTokens ns ts
    | otherwise			= formatTokens ns (t : ts)
{-    | otherwise			= error $ "ns = " ++ show (iname:ns)
				    ++ "\n" ++ replicate 50 '-'
				    ++ "\nts = " ++ show (TkProp r : ts)
-}
formatTokens (iname:ns) (t@(TkType r) : ts)
    | getName iname == "Type"	= IdTok (setRange r iname) : formatTokens ns ts
    | otherwise			= formatTokens ns (t : ts)
{-    | otherwise			= error $ "ns = " ++ show (iname:ns)
				    ++ "\n" ++ replicate 50 '-'
				    ++ "\nts = " ++ show (TkType r : ts)
-}
formatTokens ns (t@(TkId (_, "_")) : ts) =
    SimpleTok t : formatTokens ns ts
formatTokens (iname:ns) (t@(TkId (r, x)) : ts)
    | getName iname == x    = IdTok (setRange r iname) : formatTokens ns ts
    | otherwise		    = formatTokens ns (t:ts)
{-
    | otherwise		    = error $ "ns = " ++ show (iname:ns)
				    ++ "\n" ++ replicate 50 '-'
				    ++ "\nts = " ++ show (t : ts)
-}
formatTokens (iname:ns) (t@(TkOp (r, x)) : ts)
    | getName iname == x    = OpTok (setRange r iname) : formatTokens ns ts
    | x == "!"		    = SimpleTok t : formatTokens (iname:ns) ts
    | otherwise		    = formatTokens ns (t:ts)
{-    | otherwise		    = error $ "ns = " ++ show (iname:ns)
				    ++ "\n" ++ replicate 50 '-'
				    ++ "\nts = " ++ show (t : ts)
-}
formatTokens ns (tp@(TkDot _) : tn@(TkInt (_, n)) : ts)
    | n == 1 || n == 2	= map SimpleTok [tp,tn] ++ formatTokens ns ts
formatTokens ns (t@(TkInt _) : ts) =
	SimpleTok t : formatTokens (dropNat ns) ts
    where
	dropNat (x : xs)
	    | getName x == "zero"   = xs
	    | getName x == "succ"   = dropNat xs
	    | otherwise		    = error $ "Expected 'succ' or 'zero', found " ++ show x
formatTokens ns (t@(TkOpenSquare _) : ts) =
	SimpleTok t : formatTokens (dropList ns) ts
    where
	dropList (x : xs)
	    | getName x == "nil"    = xs
	    | getName x == ":"	    = dropList xs
	    | otherwise		    = error $ "Expected 'nil' or ':', found " ++ show x
formatTokens ns (t : ts)
    | isInfix t	= SimpleTok t : dropInfix ts
    where
	isInfix (TkInfixR _)	= True
	isInfix (TkInfixL _)	= True
	isInfix (TkInfix _)	= True
	isInfix _		= False

	dropInfix (t@(TkInt _) : ts)	= SimpleTok t : dropOps ts
	
	dropOps (t@(TkOp _) : ts) =
	    SimpleTok t : dropMoreOps ts
	dropOps (t0@(TkBackQuote _) : t@(TkId _) : t1@(TkBackQuote _) : ts) =
	    map SimpleTok [t0, t, t1] ++ dropMoreOps ts
	dropMoreOps (t@(TkComma _) : ts) =
	    SimpleTok t : dropOps ts
	dropMoreOps ts	= formatTokens ns ts

formatTokens ns (t:ts) = SimpleTok t : formatTokens ns ts

-- Texifying --------------------------------------------------------------

texify :: [FilePath] -> IO String
texify files =
    do	let file = last files
	ts <- parseStuff tokensParser file
	ps <- smash <$> mapM (parseStuff programParser) files
	case (ts,ps) of
	    (ParseOk _ ts, ParseOk _ ps)    ->
		do  --putStr $ unlines $ map show ts
		    let ips = runConvM $ convertPrograms ps
		    --print ips
		    --let inames = getInternalNames $ last ips
		    --putStr $ unlines $ map show $ inames
		    let fancyts = phase0 (last ips) ts
		    --putStr $ phase6 $ phase5 $ phase4 $ phase3 $ phase2 $ phase1 fancyts
		    return $ phase6 $ phase5 $ phase4 $ phase3 $ phase2 $ phase1 fancyts
		    --return $ texifyTokens fancyts
	    _	-> error "Parse error (run without -t to get a nice message)"
    where
	convertPrograms [] = return []
	convertPrograms (p:ps) =
	    uncurry (:) <$> programToInternalM p (convertPrograms ps)

	-- smash ParseResults together
	smash = foldr cat (ParseOk undefined [])
	    where
		cat (ParseOk s x) (ParseOk _ xs)    = ParseOk s (x:xs)
		cat (ParseFailed r x y) _	    = ParseFailed r x y
		cat _ (ParseFailed r x y)	    = ParseFailed r x y

	parseStuff p file
	    | isLiterate file	= parseLiterateFile p file
	    | otherwise		= parseFile p file

-- New texify algorithm ---------------------------------------------------

data Thing t	= TeX String
		| Agda [t]

instance Show t => Show (Thing t) where
    show (TeX s)    = "TeX\n" ++ unlines (map ("  "++) $ lines s)
    show (Agda ts)  = "Agda\n" ++ unlines (map ("  "++)
				    $ lines $ concatMap show ts)

newThing :: ([a] -> [b]) -> Thing a -> Thing b
newThing f (TeX s)	= TeX s
newThing f (Agda ts)	= Agda $ f ts

newThings = map . newThing

-- Phase 0: Add format directives to tokens

phase0 :: Program -> [Token] -> [FancyToken]
phase0 p = formatTokens (getInternalNames p)

-- Phase 1: Separating Agda from TeX

phase1 :: [FancyToken] -> [Thing FancyToken]
phase1 ts = map toThing1 $ chunks isTeX ts
    where
	isTeX (SimpleTok (TkTeX _)) = True
	isTeX _			    = False

	getTeX (SimpleTok (TkTeX s))	= s

	toThing1 ts@(t:_)
	    | isTeX t	= TeX $ concatMap getTeX ts
	    | otherwise	= Agda ts

-- Phase 2: Splitting Agda tokens into lines

type Col    = Int
data Line   = Line Col [FancyToken]

instance Show Line where
    show (Line c ts) = "Line " ++ show c ++ "\n" ++
			unlines (map ("  "++) $ map show ts)

phase2 :: [Thing FancyToken] -> [Thing Line]
phase2 = newThings (split 0)
    where
	split _ []	= []
	split 0 ts	= split (tokLine $ head ts) ts
	split line ts	=
	    case ts0 of
		[]  -> Line 1 [] : split (tokLine $ head ts1) ts1
		_   -> Line (tokCol $ head ts0) ts0 : split (line + 1) ts1
	    where
		(ts0,ts1) = span (\t -> tokLine t == line) ts

	tokLine t = posLine $ rStart $ getRange t
	tokCol	t = posCol  $ rStart $ getRange t

-- Phase 3: Insert explicit spaces between tokens.
--	    (a box space is inserted for indented lines).
--	    After this phase the end position of a token
--	    should equal the start position of the next
--	    token on the same line.

phase3 :: [Thing Line] -> [Thing Line]
phase3 = newThings (map insertSpace)
    where
	insertSpace (Line c ts) = Line c $ box ++ iSpace ts
	    where
		box | c == 1	= []
		    | otherwise	= [SpaceTok r Box]
		start = rStart $ getRange $ head ts
		r   = Range (start { posCol = 1 }) start

	iSpace (t1:t2:ts)
	    | getEnd t1 == getStart t2	= t1 : iSpace (t2:ts)
	    | otherwise			= t1 : SpaceTok r (Fixed 1) : iSpace (t2:ts)
	    where
		r = Range (rEnd $ getRange t1) (rStart $ getRange t2)
	iSpace [t]  = [t, SpaceTok r (Fixed 0)]
	    where
		end = rEnd $ getRange t
		r = Range end end
	iSpace []   = []

	getStart    = posCol . rStart . getRange
	getEnd	    = posCol . rEnd . getRange

-- Phase 4: Turn the lines into a tree shape

-- | The parent of a line @l@ is the closest line above @l@
--   that is indented less than @l@. Empty lines doesn't count.
data LineTree = TreeLine Col [FancyToken] [LineTree]

instance Show LineTree where
    show (TreeLine c ts ls)
	= "Line " ++ show c ++ "\n" ++
	    unlines (map ("  "++) $ map show ts) ++ "\n" ++
	    unlines (map ("  "++) $ lines $ unlines $ map show ls)

phase4 :: [Thing Line] -> [Thing LineTree]
phase4 = newThings buildTree
    where
	buildTree [] = []
	buildTree (Line c [] : ls) = TreeLine c [] [] : buildTree ls
	buildTree (Line c ts : ls) =
		TreeLine c ts (buildTree ls0) : buildTree ls1
	    where
		(ls0, ls1) = span (\l -> isEmpty l || getIndent l > c) ls

	isEmpty (Line _ ts) = null ts

	getIndent (Line c _) = c

-- Phase 5: Computing line indentations and boxes

data BoxedLine	= BLine [[FancyToken]] [FancyToken] [Child]
type BoxIndex	= Int
type Indent	= Int
data Child	= Child BoxIndex Indent BoxedLine

instance Show Child where
    show (Child n i l) = "box" ++ show n ++ "+" ++ show i ++ " -> " ++ show l

instance Show BoxedLine where
    show (BLine boxs ts ls) =
	    "BLine\n" ++
		unlines (map ("  "++) $ concatMap showBox boxs) ++
		"  rest\n" ++
		unlines (map ("    "++) $ map show ts) ++
		unlines (map ("  "++) $ lines $ unlines $ map show ls)
	where
	    showBox ts = "box" : map ("  "++) (map show ts)

phase5 :: [Thing LineTree] -> [Thing BoxedLine]
phase5 = newThings (map computeBoxes)
    where

	computeBoxes (TreeLine _ ts [])	= BLine [] ts []
	computeBoxes (TreeLine c ts ls) =
		BLine boxs ts' $ zipWith mkChild nboxs ls
	    where
		is	= map getIndent ls

		mkChild (n,i) l = Child n i $ computeBoxes l

		-- the box indices for the children
		nboxs	= map findBox is

		findBox c = findBox' c $ zip [0..] ws
		    where
			findBox' c []		    = (-1, c - 1)
			findBox' c ((n, w):_)
			    | w > c		    = (n - 1, c - 1)
			findBox' c [(n,w)]	    = (n, c - w)
			findBox' c ((n, w) : boxs@((_,w') : _))
			    | w' > c	= (n, w - c)
			    | otherwise	= findBox' c boxs

		-- is is decreasing, which means that
		-- cs will be strictly increasing.
		cs  = map head $ group $ reverse is
		tss = splitTokens cs ts
		(wboxs,ts') = getBoxes tss
		(ws, boxs) = unzip wboxs

	-- Compute the boxes and their width from the chopped up token list
	getBoxes :: [[FancyToken]] -> ([(Int, [FancyToken])], [FancyToken])
	getBoxes [ts] = ([], ts)
	getBoxes (box:ts@((t:_):_))   = ((w,box):boxs, ts')
	    where
		w   = posCol $ rStart $ getRange t
		(boxs, ts') = getBoxes ts
	getBoxes tss	= error $ show tss

	getIndent (TreeLine c _ _) = c

	-- Split a list of tokens at given columns
	-- Should only produce the empty list as the last element of the result.
	splitTokens [] ts = [ts]
	splitTokens (c:cs) ts = ts0 `cons` splitTokens cs ts1
	    where
		cons [] xss = xss
		cons xs xss = xs : xss

		(ts0,ts1) = span (endsBefore c) ts
		
				-- The eol token shouldn't be included in a box
		endsBefore c (SpaceTok r (Fixed 0)) = False
		endsBefore c t = posCol (rEnd $ getRange t) <= c

	getNextIndents = getIndents maxBound
	    where
		getIndents prev (Line _ [] : ls) = getIndents prev ls
		getIndents _ (Line 1 _ : _) = []
		getIndents prev (Line c _ : ls)
		    | c < prev	= c : getIndents c ls
		    | otherwise	= [c]
		getIndents _ [] = []

-- Phase 6: Generating TeX

phase6 :: [Thing BoxedLine] -> String
phase6 = unlines . concatMap thingToTeX
    where
	thingToTeX (TeX s)	= [s]
	thingToTeX (Agda ls)	=
	    [ beginAgda ]
	    ++ concatMap blineToTeX ls ++
	    [ endAgda ]
	
	blineToTeX (BLine [] [] []) =
	    [ mathMode "{}" ++ par ]
	blineToTeX (BLine boxs ts ls) =
		[ "{" ]
		++ map ("  "++) (zipWith setbox [0..] boxs'
		    ++ [ mathMode (copy ++ tokensToTeX ts') ++ par ]
		    ++ concatMap childToTeX ls
		    )
		++ [ "}" ]
	    where
		copy = case length boxs of
			    0	-> ""
			    n	-> copyBox (n - 1)

		setbox 0 box = setBox 0 $ mathMode $ tokensToTeX box
		setbox n box = setBox n	$ mathMode
					$ copyBox (n - 1) ++
					  tokensToTeX box

		(boxs',ts') = initLast $ mathSpace (boxs ++ [ts])

{-
		box' = mathSpace box
		ts' | null box		    = mathSpace' ts
		    | isMathRel (last box)  = mathSpace ts
		    | otherwise		    = mathSpace' ts
-}
	childToTeX (Child n i l) = saveWidth n i : blineToTeX l

	tokensToTeX ts = concatMap tokString ts

	isSpace (SpaceTok _ _)   = True
	isSpace _		    = False

	initLast [x] = ([], x)
	initLast (x:xs) = (x : xs', y)
	    where (xs',y) = initLast xs

	headCons x (xs:xss) = (x:xs) : xss

	mathSpace tss = skipIndent tss
	    where
		skipIndent ((t:ts):tss)
		    | isSpace t	    = t `headCons` mathSpace' (ts:tss)
		    | otherwise	    = mathSpace' ((t:ts):tss)

	mathSpace' ([]:tss) = [] : mathSpace' tss
	mathSpace' ([t]:(t':ts):tss)
	    | isSpace t && isMathRel t'	= [] : mathSpace' ((t':ts):tss)
	    | isMathRel t && isSpace t'	= [t] : mathSpace' (ts:tss)
	    | otherwise			= [t] : mathSpace' ((t':ts):tss)
	mathSpace' ((t:t':ts):tss)
	    | isSpace t && isMathRel t'	= mathSpace' ((t':ts):tss)
	    | isMathRel t && isSpace t'	= mathSpace' ((t:ts):tss)
	    | otherwise			= t `headCons` mathSpace' ((t':ts):tss)
	mathSpace' [[t]]    = [[t]]
	mathSpace' []	    = []
	mathSpace' tss	    = error $ show tss

-- Ye olde algorithm ------------------------------------------------------

{-
data St = St	{ stRow	    :: Int
		, stCol	    :: Int
		, stEndCol  :: Int
		, stLayout  :: [Int]
		, stTeXMode :: Bool
		, stPrevTok :: FancyToken
		}

initSt = St 0 1 0 [1] True (SimpleTok $ TkTeX "")

type M = StateT St (Writer String)

runM = execWriter . flip evalStateT initSt

texifyTokens :: [FancyToken] -> String
texifyTokens ts = runM (texToks ts)

getIndent :: St -> Int -> Int
getIndent s c = c - l
    where l:_ = stLayout s

getSpace :: FancyToken -> M String
getSpace t = theSpace <$> get
    where
	r   = getRange t
	row = posLine (rStart r)
	col = posCol (rStart r)

	newlines s = replicate (min 2 (row - stRow s)) newline

	leftarg s
	    | isMathRel t   = "{}"
	    | otherwise	    = ""

	horSpace s
	    | isMathRel t		= ""
	    | isMathRel (stPrevTok s)	= ""
	    | otherwise			= singleSpace

	theSpace s
	    | stRow s == 0	= ""
	    | stRow s /= row    = concat (newlines s) ++
				  longSpace (getIndent s col) ++
				  leftarg s
	    | col /= stEndCol s = horSpace s
	    | otherwise		= ""


putSpace :: FancyToken -> M ()
putSpace t = tell =<< getSpace t

push :: Range -> M ()
push r = modify $ \s -> s { stLayout = posCol (rStart r) : stLayout s }

pop :: M ()
pop = modify $ \s -> s { stLayout = tail $ stLayout s }

step :: FancyToken -> M ()
step t = modify $ \s -> s   { stRow	= row
			    , stCol	= col
			    , stEndCol	= end
			    , stPrevTok = t
			    }
    where
	r   = getRange t
	row = posLine (rStart r)
	col = posCol (rStart r)
	end = posCol (rEnd r)

reset :: M ()
reset = put initSt

enterAgdaMode :: M ()
enterAgdaMode =
    do	s <- get
	if stTeXMode s
	    then do tell beginAgda
		    modify $ \s -> s { stTeXMode = False }
	    else return ()

exitAgdaMode :: M ()
exitAgdaMode =
    do	s <- get
	if stTeXMode s
	    then return ()
	    else do tell endAgda
		    reset

texToks :: [FancyToken] -> M ()
texToks (SimpleTok (TkTeX s) : ts) =
    do	exitAgdaMode
	tell s
	reset
	texToks ts
texToks (SimpleTok TkEOF : _) = exitAgdaMode
texToks []	    = exitAgdaMode
texToks ts =
    do	enterAgdaMode
	texToks' ts
    where
	texToks' ts = case ts of
	    t@(SimpleTok (TkVOpenBrace r)) : t' : ts ->
		do  putSpace t
		    when (posCol (rStart $ getRange t') /= 1) $
			do  push (getRange t')
			    tell "\\hsalign{"
		    step t
		    texToks (t' : ts)
	    t@(SimpleTok (TkVCloseBrace r)) : ts    ->
		do  pop
		    tell "}"
		    putSpace t
		    step t
		    texToks ts
	    t : ts		    ->
		do  putSpace t
		    tell (tokString t)
		    step t
		    texToks ts
-}
isMathRel t =
    case t of
	SimpleTok (TkArrow _)	    -> True
	SimpleTok (TkDblColon _)    -> True
	SimpleTok (TkEqual _)	    -> True
	SimpleTok (TkOp _)	    -> True
	_			    -> False

tokString (SpaceTok _ Box) = boxSpace
tokString (SpaceTok _ (Fixed n)) = concat $ replicate n singleSpace
tokString (IdTok n) =
    case n of
	IDef _ x	-> definition $ escapeName x
	IVar _ x	-> variable $ escapeName x
	ISort _ x	-> newsort $ escapeName x
tokString (OpTok n) =
    case n of
	IDef _ x	-> definition $ escapeOp x
	IVar _ x	-> variable $ escapeOp x
	ISort _ x	-> newsort $ escapeOp x
tokString (SimpleTok t) =
    case t of
	TkLambda _	    -> "\\lambda{}"
	TkArrow _	    -> "{}\\to{}"
	TkId (_,x)	    -> escapeName x
	TkLParen _	    -> "("
	TkRParen _	    -> ")"
	TkSemi _	    -> ";"
	TkVSemi _	    -> ""
	TkComma _	    -> ","
	TkUniverse (_,n)    -> newsort $ "\\#" ++ show n
	TkInfix _	    -> keyword "infix"
	TkInfixL _	    -> keyword "infixl"
	TkInfixR _	    -> keyword "infixr"
	TkBackQuote _	    -> "`"
	TkWhere _	    -> keyword "where"
	TkData _	    -> keyword "data"
	TkSort _	    -> keyword "sort"
	TkInt (_,n)	    -> show n
	TkDblArrow _	    -> "{}\\Rightarrow{}"
	TkOpenBrace _	    -> "\\{"
	TkCloseBrace _	    -> "\\}"
	TkOpenSquare _	    -> "["
	TkCloseSquare _	    -> "]"
	TkEqual _	    -> "{}\\equiv{}"
	TkDblColon _	    -> "{}\\in{}"
	TkDot _		    -> "."
	TkLet _		    -> keyword "let"
	TkIn _		    -> keyword "in"
	TkPlugin (_,x)	    -> keyword $ x ++ "{-}plugin"
	TkUnderscore _	    -> "\\_"
	TkOp (_, x)	    -> escapeOp x
	TkVOpenBrace _	    -> ""
	TkVCloseBrace _	    -> ""
	_		    -> error $ show t

keyword x	= "\\AgdaKeyword{" ++ x ++ "}"

variable x	= "\\AgdaVariable{" ++ x ++ "}"
definition x	= "\\AgdaDefinition{" ++ x ++ "}"
newsort x	= "\\AgdaSort{" ++ x ++ "}"

specialOps =
    [ "<="	|-> "{}\\leq{}"
    , "/="	|-> "{}\\neq{}"
    , ">="	|-> "{}\\geq{}"
    , "==>"	|-> "{}\\Longrightarrow{}"
    , "<=="	|-> "{}\\Longleftarrow{}"
    , "/\\"	|-> "{}\\wedge{}"
    , "\\/"	|-> "{}\\vee{}"
    , "<=>"	|-> "{}\\Leftrightarrow{}"
    , "<==>"	|-> "{}\\Longleftrightarrow{}"
    , "-->"	|-> "{}\\longrightarrow{}"
    , "<--"	|-> "{}\\longleftarrow{}"
    , "<-->"	|-> "{}\\longleftrightarrow{}"
    , "<->"	|-> "{}\\leftrightarrow{}"
    , "<-"	|-> "{}\\leftarrow{}"
    , "|-"	|-> "{}\\vdash{}"
    , "-|"	|-> "{}\\dashv{}"
    , "|="	|-> "{}\\models{}"
    , "<>"	|-> "{}\\Diamond{}"
    , "<<"	|-> "{}\\ll{}"
    , ">>"	|-> "{}\\gg{}"
    , "~>"	|-> "{}\\leadsto{}"
    , "|->"	|-> "{}\\mapsto{}"
    , "|-->"	|-> "{}\\longmapsto{}"
    , "||"	|-> "{}\\parallel{}"
    , "|><|"	|-> "{}\\Join{}"
    ]
    where (|->) = (,)

escapeOp x = maybe (mkOp x) id (lookup x specialOps)
    where
	mkOp x = "{}\\mathrel{" ++
		    concat (narrow $ intersperse "\\mkern-4mu" $ map esc x)
		    ++ "}{}"

	esc '&'	    = "\\&"
	esc '%'	    = "\\%"
	esc '\\'    = "\\backslash{}"
	esc '$'	    = "\\$"
	esc c	    = [c]

	narrow (_:"|":xs)   = "|" : narrow xs
	narrow (_:"!":xs)   = "!" : narrow xs
	narrow (_:"'":xs)   = "'" : narrow xs
	narrow (x:xs)	    = x : narrow xs
	narrow []	    = []

specialNames =
    [ "alpha"	|-> "\\alpha{}"
    , "beta"	|-> "\\beta{}"
    , "gamma"	|-> "\\gamma{}"
    , "delta"	|-> "\\delta{}"
    , "epsilon" |-> "\\epsilon{}"
    , "zeta"	|-> "\\zeta{}"
    , "eta"	|-> "\\eta{}"
    , "theta"	|-> "\\theta{}"
    , "iota"	|-> "\\iota{}"
    , "kappa"	|-> "\\kappa{}"
    , "mu"	|-> "\\mu{}"
    , "nu"	|-> "\\nu{}"
    , "xi"	|-> "\\xi{}"
    , "pi"	|-> "\\pi{}"
    , "rho"	|-> "\\rho{}"
    , "sigma"	|-> "\\sigma{}"
    , "tau"	|-> "\\tau{}"
    , "upsilon" |-> "\\upsilon{}"
    , "phi"	|-> "\\phi{}"
    , "chi"	|-> "\\chi{}"
    , "psi"	|-> "\\psi{}"
    , "omega"	|-> "\\omega{}"
    , "Gamma"	|-> "\\Gamma{}"
    , "Delta"	|-> "\\Delta{}"
    , "Theta"	|-> "\\Theta{}"
    , "Lambda"	|-> "\\Lambda{}"
    , "Xi"	|-> "\\Xi{}"
    , "Pi"	|-> "\\Pi{}"
    , "Sigma"	|-> "\\Sigma{}"
    , "Upsilon" |-> "\\Upsilon{}"
    , "Phi"	|-> "\\Phi{}"
    , "Psi"	|-> "\\Psi{}"
    , "Omega"	|-> "\\Omega{}"
    ]
    where
	(|->) = (,)

escapeName x = maybe (concatMap esc x) id (lookup x specialNames)
    where
	esc '_' = "\\_"
	esc c	= [c]

setBox n s  = "\\setbox" ++ show n ++ "=\\hbox{" ++ s ++ "}"
copyBox	n   = "\\copy" ++ show n
saveWidth (-1) i    = "\\edef\\Indent{" ++
			longSpace i ++ "}"
saveWidth n i	    = "\\edef\\Indent{\\hskip\\wd" ++ show n ++
			longSpace i ++ "}"
boxSpace    = "\\Indent"
mathMode "" = "${}$"
mathMode s  = "$" ++ s ++ "$"
par	    = "\\par"

beginAgda = "\\begin{agda}"
endAgda	    = "\\end{agda}"
longSpace 0 = ""
longSpace n = "\\hskip" ++ show (2 * n) ++ "mm"
newline = "{}\\\\\n"
singleSpace = "\\,"

-- HasRange Token

instance HasRange Token where
    getRange t = case t of
	TkLambda r	    -> r
	TkArrow r	    -> r
	TkId (r, _)	    -> r
	TkLParen r	    -> r
	TkRParen r	    -> r
	TkSemi r	    -> r
	TkVSemi r	    -> r
	TkComma r	    -> r
	TkUniverse (r,_)    -> r
	TkSet r		    -> r
	TkProp r	    -> r
	TkType r	    -> r
	TkInfix r	    -> r
	TkInfixL r	    -> r
	TkInfixR r	    -> r
	TkBackQuote r	    -> r
	TkWhere r	    -> r
	TkData r	    -> r
	TkSort r	    -> r
	TkInt (r, _)	    -> r
	TkDblArrow r	    -> r
	TkOpenBrace r	    -> r
	TkCloseBrace r	    -> r
	TkOpenSquare r	    -> r
	TkCloseSquare r	    -> r
	TkVOpenBrace r	    -> r
	TkVCloseBrace r	    -> r
	TkEqual r	    -> r
	TkDblColon r	    -> r
	TkDot r		    -> r
	TkLet r		    -> r
	TkIn r		    -> r
	TkPlugin (r,_)	    -> r
	TkUnderscore r	    -> r
	TkOp (r, _)	    -> r
	TkEOF		    -> noRange

