
module Plugins.FOL.Options where

import Control.Monad.Error
import Data.Char
import System.Console.GetOpt

import Interface.Options (Flag)
import Utils.Monads

data Way = SantaOtter | SantaTSTP | Otter
    deriving (Show, Eq)

data FOLOptions =
    FOLOptions	{ optTimeout	    :: Int
		, optProver	    :: Maybe String
		, optFinder	    :: Maybe String
		, optSanta	    :: Maybe String
		, optWay	    :: Way
		, optKeepFiles	    :: Bool
		, optCounterModels  :: Bool
		, optBinaryApp	    :: Bool
		, optModelSize	    :: Int
		}

folFormatFlag fmt o =
    case map toLower fmt of
	"tstp"	->
	    case optWay o of
		Otter   -> fail "cannot use '--fol-format=tstp' with '--fol-no-santa'"
		_	-> return $ o { optWay = SantaTSTP }
	"otter"	->
	    case optWay o of
		SantaTSTP   -> return $ o { optWay = SantaOtter }
		_	    -> return o
	fmt	-> fail $ "Unrecognized fol format: '" ++ fmt ++ "'"

noSantaFlag o =
    case optWay o of
	SantaTSTP   -> fail "cannot use '--fol-format=tstp' with '--fol-no-santa'"
	_	    -> return $ o { optWay = Otter }

folSantaFlag s o    = return $ o { optSanta	    = Just s }
folFinderFlag p o   = return $ o { optFinder	    = Just p }
folProverFlag p o   = return $ o { optProver	    = Just p }
keepFilesFlag o	    = return $ o { optKeepFiles	    = True }
counterModelsFlag o = return $ o { optCounterModels = True }
binaryAppFlag o	    = return $ o { optBinaryApp	    = False }
modelSizeFlag s o   =
    do  n <- readM s
	return o { optModelSize = n }
    `catchError`
	\_ -> fail "option '--fol-model-size requires an int argument\n"
timeoutFlag s o	    =
    do  n <- readM s
	return o { optTimeout = n }
    `catchError`
	\_ -> fail "option '--fol-timeout' requires an int argument\n"

defaultOptions :: FOLOptions
defaultOptions =
    FOLOptions	{ optTimeout	    = 3000
		, optProver	    = Nothing
		, optFinder	    = Nothing
		, optSanta	    = Nothing
		, optKeepFiles	    = False
		, optCounterModels  = False
		, optBinaryApp	    = True
		, optModelSize	    = 8
		, optWay	    = SantaOtter
		}

options :: [OptDescr (Flag FOLOptions)]
options =
    [ Option [] ["timeout"]	(ReqArg timeoutFlag "N")
		"set timeout for first order prover to N milliseconds"
    , Option []	["prover"]	(ReqArg folProverFlag "PROVER")
		"use PROVER as the first order prover"
    , Option [] ["model-finder"] (ReqArg folFinderFlag "FINDER")
		"use the model finder FINDER to find counter models"
    , Option []	["santa"]	(ReqArg folSantaFlag "SANTA")
		"use SANTA as santa"
    , Option [] ["counter-models"] (NoArg counterModelsFlag)
		"search for counter models if proving fails"
    , Option [] ["model-size"] (ReqArg modelSizeFlag "N")
		"set maximum counter model size to N (default=8)"
    , Option [] ["no-santa"]	(NoArg noSantaFlag)
		"don't use santa to generate input to prover"
    , Option [] ["format"]	(ReqArg folFormatFlag "FMT")
		"send first order goals in format FMT [otter|tstp]"
    , Option []	["keep-files"]	(NoArg keepFilesFlag)
		"do not remove intermediate files"
    , Option [] ["no-binary-app"]   (NoArg binaryAppFlag)
		"use function symbols instead of binary application whenever possible"
    ]

