
module Plugins.FOL.Pretty where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.State

import Syntax.Position
import Syntax.Name
import Syntax.Pretty
import Utils.Names
import Utils.Map hiding (map)

import Plugins.FOL.Types
import Plugins.FOL.Primitive
import Plugins.FOL.Constants

-- Find all constants in a formula ----------------------------------------

class HasConstants a where
    constants :: a -> [String]

instance (HasConstants a, HasConstants b) => HasConstants (a,b) where
    constants (x,y) = constants x ++ constants y

instance HasConstants a => HasConstants [a] where
    constants xs = concatMap constants xs

instance HasConstants Formula where
    constants TRUE      = []
    constants FALSE     = []
    constants (And p q) = constants (p,q)
    constants (Or  p q) = constants (p,q)
    constants (Not p)   = constants p
    constants (Implies  p q) = constants (p,q)
    constants (Equiv p q) = constants (p,q) 
    constants (ForAll _ p) = constants (p $ FOLVar "_")
    constants (Exists _ p) = constants (p $ FOLVar "_")
    constants (Predicate x ts)	= x : constants ts

instance HasConstants FOLTerm where
    constants (FOLFun x ts) = x : constants ts
    constants (FOLVar _)    = []

-- Pretty printing FOL propositions ---------------------------------------

-- initialNames :: Formula -> UsedNames
-- initialNames p = foldl (\ m x -> insert x 0 m) noNamesUsed $ folConstants ++ constants p
  --keywords

(  precTop
 : precEquiv
 : precImplies
 : precOr
 : precAnd
 : precEqual
 : precNot
 : precInfix
 : precAppL
 : precAppR
 : precExists
 : precForAll
 : _
 ) = [0..]

precMax = precAppR

-- indentation is 2 (nest)

isOp x = isOperator (Name noRange x 0)

instance Pretty FOLTerm where
    prettyPrec n (FOLVar x) = text x
    prettyPrec n (FOLFun x []) = text x
    prettyPrec n (FOLFun xProj [t,t']) | xProj == kwProj =
	    prettyPrec precMax t <> text "." <> prettyProj t'
	where
	    prettyProj (FOLFun p [])
		| p == kwFst	= text "1"
		| p == kwSnd	= text "2"
		| otherwise	= text p
	    prettyProj _ = error "impossible: Plugins.FOL.Pretty.prettyProj"
    prettyPrec n (FOLFun xApp [FOLFun xApp' [FOLFun op [],t],t'])
	| all (kwApp ==) [xApp, xApp'] && isOp op =
	    mparens (n > precInfix) $
		fsep [ prettyPrec (precInfix + 1) t
		     , text op
		     , prettyPrec (precInfix + 1) t'
		     ]
    prettyPrec n (FOLFun xApp [t,t'])
	| xApp == kwApp = 
	    mparens (n > precAppL) $ 
                fsep [ prettyPrec precAppL t
                     , nest 2 $ prettyPrec precAppR t'
                     ]
    prettyPrec n (FOLFun op [t,t'])
	| isOp op	= 
	    mparens (n > precInfix) $
		fsep [ prettyPrec (precInfix + 1) t
		     , text op
		     , prettyPrec (precInfix + 1) t'
		     ]
    prettyPrec n (FOLFun x ts) = 
        mparens (n>precAppL) $ 
                fsep $ text x : map (nest 2 . prettyPrec precAppR) ts

-- TODO : Lamport-style of printing formulas
-- Nested Forall/Exists needs better layout

instance Pretty Formula where
    prettyPrec n p = flip evalState noNamesUsed $ prettyFM n p -- hack

prettyFM :: Int -> Formula -> State UsedNames Doc
prettyFM n TRUE      = return $ text fol_True
prettyFM n FALSE     = return $ text fol_False
prettyFM n (And p q) = do dp <- prettyFM precAnd p
                          dq <- prettyFM precAnd q
                          return $ mparens (n>precAnd) $
                            fsep [ dp
                                 , nest 2 $ text fol_And
                                 , nest 2 dq
                                 ]
prettyFM n (Or  p q) = do dp <- prettyFM precOr p
                          dq <- prettyFM precOr q
                          return $ mparens (n>precOr) $
                            fsep [ dp
                                 , nest 2 $ text fol_Or
                                 , nest 2 dq
                                 ]
prettyFM n (Not p)   = do dp <- prettyFM precNot p
                          return $ mparens (n>precNot) $
                            fsep [ text fol_Not
                                 , nest 2 $ dp
                                 ]
prettyFM n (Implies  p q) = 
    do dp <- prettyFM precImplies p
       dq <- prettyFM precImplies q
       return $ mparens (n>precImplies) $
                            fsep [ dp
                                 , nest 2 $ text fol_Implies
                                 , nest 2 dq
                                 ]
prettyFM n (Equiv p q) = 
    do dp <- prettyFM precEquiv p
       dq <- prettyFM precEquiv q
       return $ mparens (n>precEquiv) $
                            fsep [ dp
                                 , nest 2 $ text fol_Equiv
                                 , nest 2 dq
                                 ]
prettyFM n (ForAll x p) = 
    do name <- fresh x
       dp <- prettyFM precForAll (p (FOLVar $ mkName name))
       return $ fsep $ (if (n==precForAll) then [] else [text fol_ForAll]) ++
                  [ nest 2 $ text $ mkName name
                  , nest 2 $ dp
                  ]
prettyFM n (Exists x p) = 
    do name <- fresh x
       dp <- prettyFM precExists (p (FOLVar $ mkName name))
       return $ fsep $ (if (n==precExists) then [] else [text fol_Exists]) ++
                  [ nest 2 $ text $ mkName name
                  , nest 2 $ dp
                  ]
prettyFM n (Predicate xEqual [t,t']) | xEqual == kwEqual = 
    return $ mparens (n>precEqual) $ 
                fsep [ prettyPrec precTop t
                     , text "=="
                     , prettyPrec precTop t'
                     ]
prettyFM n (Predicate op [t, t']) | isOp op =
    return $ mparens (n > precInfix) $
		fsep [ prettyPrec (precInfix + 1) t
		     , text op
		     , prettyPrec (precInfix + 1) t'
		     ]
prettyFM n (Predicate p ts) = do let dts = map (prettyPrec precAppR) ts
                                 return $ mparens (n>precAppL) $
                                        fsep $ (text p) :
                                               map (nest 2) dts

mkName (Name _ x n) = x ++ "_" ++ show n
mkName DontCare	    = error $ "impossible: Plugins.FOL.Pretty.mkName DontCare"

instance Show FOLTerm where
    show = show . pretty

instance Show Formula where
    show = show . pretty

